;-----------------------------------------------------------------------------+
;                                                                             |
;                       SKIN.lsp   Version 1.0   7/5/88                       |
;                                                                             |
;                                 Larry Knott                                 |
;                                                                             |
;    SKINning is the process of taking an OUTLINE and "pushing" it along a    |
;    3D curve (or PATH).  This program creates 3DMESH entities as it pushes   |
;    an OUTLINE along a PATH.  PATHs or OUTLINEs can be defined with the      |
;    following entities: LINEs, ARCs, PLINEs or 3DPOLYs.                      |
;                                                                             |
;    The user has control of the density (or number of SEGMENTs) ALONG and    |
;    AROUND the SKIN generated by means of the initial prompt:                |
;                                                                             |
;         CURRENT SETTINGS:  1.0000 distance along arc per segment,           |
;         4 segments minimum for 90 degree arc,  12 segments around outline.  |
;         Do you want to change these settings? <N>: y                        |
;                                                                             |
;         >>Distance along arc per segment:                                   |
;         >>Minimum # segments for 90 deg. arc:                               |
;         >>Number of segments around outline:                                |
;                                                                             |
;      1) The distance along ARC per segment.                                 |
;            This controls the minimum number of segments generated ALONG     |
;            ARC segments on the PATH by distance.                            |
;      2) The minimum number of segments for a 90 degree ARC.                 |
;            This defines the minimum number of segments to be generated      |
;            ALONG an ARC segment with a 90 degree included angle.            |
;      3) The number of segments AROUND the OUTLINE.                          |
;            The number of segments to generate around the outline as it is   |
;            pushed along the PATH.                                           |
;                                                                             |
;    Useful defaults are provided.  Once values have been defined, or the     |
;    defaults taken, they are saved for that editing session.  The next       |
;    prompt is:                                                               |
;                                                                             |
;         Select path:                                                        |
;                                                                             |
;    This is the PATH that the OUTLINE will be "pushed" along.  Select any    |
;    LINE, ARC, PLINE or 3DPOLY.  The next prompt is for the OUTLINE.  If     |
;    an outline has previously been defined in the drawing, the message:      |
;                                                                             |
;         >>The outline already exists.  Redefine it? Yes/<No>:               |
;                                                                             |
;    is displayed.  If "Yes" is supplied to the above prompt or no outline    |
;    existed previously, you will be asked:                                   |
;                                                                             |
;         Select outline:                                                     |
;         Insertion base point:                                               |
;                                                                             |
;    SKIN will then generate 3DMESHes defined by the PATH and OUTLINE         |
;    selected.                                                                |
;                                                                             |
;                                                                             |
;    Note:                                                                    |
;      >  If the PATH contains ARC segments that are "flat" in the currect    |
;         VPOINT, SKIN may generate the 3DMESH in the wrong direction.        |
;         Change the vpoint to correct this.                                  |
;                                                                             |
;-----------------------------------------------------------------------------+

(princ "\nLoading SKIN.lsp...")

;---------------------------- CREATE NEW *ERROR( -----------------------------|

(defun skin-er (n)
(command "undo" "end" "undo" "1")
(moder)   ;  UNDO WILL RESTORE SETVARS.
(if (eq n "Function cancelled")
  (princ "\nFunction cancelled by user.  Skin terminated.")
  (princ "\nError: Please report problem to LARRYK on 3mail."))
(setq *error* lisp-er)
(prin1))

;------------------------------- SAVE OLD MODES ------------------------------|

(defun getset (a)                                            ;RECURSIVE
(setq mlst '())
(foreach n a
  (setq mlst (append mlst (list (list (car n) (getvar (car n)))))))
(foreach n a (setvar (car n) (cadr n)))
(command "undo" "group")
(setq orgucs (if (> (strlen (getvar "ucsname")) 0) (getvar "ucsname") nil))
(if (or (not orgucs) (eq orgucs "$$skin"))
  (if (tblsearch "ucs" "$$skin")
     (command "ucs" "s" "$$skin" "y")
     (command "ucs" "s" "$$skin")))    ;  EXPERT HAS NO EFFECT ON REDEFINTITION.
)

;----------------------------- RESTORE OLD MODES -----------------------------|

(defun moder ()
(command "ucs" "r" (if orgucs orgucs "$$skin"))
(foreach n mlst (setvar (car n) (cadr n))))

;------------------------------ COMMON FUNCTIONS -----------------------------|
;
;    REMOVE ENTGET IN GETVAL.  SEND ENTITY LIST AROUND, NOT ENAME.  LESS
;    OVERHEAD.

(defun getval (n e) (cdr (assoc n (entget e))))
(defun 70bit? (n e) (if (zerop (logand n (getval 70 e))) nil T))
(defun midp (l k / ll kk) (mapcar '(lambda (ll kk) (/ (+ ll kk) 2)) l k))

;---------------------------- CHANGE SKIN SETTINGS ---------------------------|
;
;    SHOW DEFAULTS, ADD INITGET, MAKE MORE ROBUST.

(defun cngset ()
(setq arcdst  (if arcdst arcdst 1.0)
      arcmin  (if arcmin arcmin 16))
(princ (strcat "\n\nCURRENT SETTINGS:  " (rtos arcdst) " distance along arc "
  "per segment,\n" (itoa (/ arcmin 4)) " segments minimum for 90 degree arc,"
  "  " (itoa (getvar "surftab2")) " segments around outline."))
(if (/= (getstring "\nDo you want to change these settings?  Yes/<No>: ") "")
  (progn   (textscr)
  (setq    arcdst    (getreal "\n>>Distance along arc per segment: "))
  (setq    arcmin (* (getint "\n>>Minimum # segments for 90 deg. arc: ") 4))
  (setvar "surftab2" (getint "\n>>Number of segments around outline: ")))))

;--------------------------------- GET ENTITY --------------------------------|

(defun getent (t1 t2 retry / e0 e1)
(setq e0 T)
(while e0
  (if (setq e0 (entsel (strcat "\nSelect " t2 ": ")))
    (if (member (cdr (assoc 0 (setq e1 (entget (car e0))))) t1)
      (setq e0 nil)
      (princ (strcat " " (cdr (assoc 0 e1)) ", Not a valid entity.")))
    (setq e0 (if retry (princ "  No entity found.") nil)))
) (cdr (assoc -1 e1)))                                    ;  SUPPLY ENAME ONLY

;------------------------------ DEFINE OUTLINE -------------------------------|

(defun outlin (/ p1 t1 e1)
(setq t1
  (cond
     ((tblsearch "BLOCK" "$$skin")
        (princ "\n>>The outline already exists.  ")
        (initget "Yes No")
        (if (= (getkword "Redefine it?  Yes/<No>: ") "Yes") T nil))
     (T)))
(cond
  (t1 (setq e1 (getent '("POLYLINE" "LINE" "ARC" "CIRCLE") "outline" T))
      (setq p1 (getpoint "\nInsertion base point: "))
      (command "layer" "n" "$$skin" "off" "$$skin" ""
               "copy" e1 "" '(0 0 0) '(0 0 0)
               "chprop" (entlast) "" "la" "$$skin" ""
               "block" "$$skin" p1 (entlast) "")
      (redraw e1))))

;----------------------------- DRAW LINE ROUTINE -----------------------------|
;
; BUG IN 2.5 RAMDOM START POINTS FOR TABSURF'S, USE VS RULESURF WHEN FIXED!
;(command "ucs" "p" "line" p1 p2 ""
;         "tabsurf"  (list t1 p1) (list (setq t2 (entlast)) p2))

(defun drwlin (e / p1 p2 t1 t2) ;  HAVING E1 E2 LOCAL -> LOSE UCS
(setvar "surftab1" (getvar "surftab2"))
(setq p1 (getval 10 e)
      p2 (cond ((getval 11 e))              ;  LINE.
               ((getval 10 (entnext e)))))  ;  POLYLINE.
(if notwcs (setq p1 (trans p1 path 1)
                 p2 (trans p2 path 1)))
(command "ucs" "za" p1 p2                   ;  RANDOM XY PLANE. CHANGE
         "insert" "*$$skin" '(0 0 0) 1 0)
(setq t1 (entlast))                         ;  WATCH DIST. BUG.
(command "insert" "*$$skin" (list 0 0 (distance p1 p2)) 1 0
         "ucs" "p"
         "rulesurf" (list t1 p1) (list (setq t2 (entlast)) p2))
(entdel t1) (entdel t2))

;------------------------------ DRAW ARC ROUTINE -----------------------------|
;
;    WHEN THE CURRENT VIEWDIR IS PARALELL TO THE AXIS OF REVOLUTION, THE
;    REVSURF OR TABSURF RESULTS ARE NOT CONSISTANT. CHANGE VPOINT.

(defun drwarc (e)
(setq c1 (getval 10 e)
      c2 (mapcar '+ c1 '(0 0 1))
      r1 (getval 40 e)
      a1 (getval 50 e)
      a2 (getval 51 e)
      ia (abs (if (> a1 a2) (- a1 (* pi 2) a2) (- a2 a1)))  ;  NEEDS REFINEMENT.
;       p1 (polar c1 a1 r1)     BUG- POLAR OF 3D SHOULD RETURN 3D PT.
;       p2 (polar c1 a2 r1)
      p1 (append (polar c1 a1 r1) '(0))
      p2 (append (polar c1 a2 r1) '(0)))
(setvar "surftab1" (abs (fix (if (> (fix (/ (* 2 pi r1) arcdst)) arcmin)
  (/ (* ia r1) arcdst) (* arcmin (/ ia pi))))))
(setq p1 (trans p1 e 1)
      p2 (trans p2 e 1)
      c1 (trans c1 e 1)
      c2 (trans c2 e 1))
(command "ucs" "3p" p1 c1 c2
         "insert" "*$$skin" '(0 0 0) 1 0
         "ucs" "p")
(setq e1 (entlast))
(command "line" c1 c2 ""
         "revsurf" (list e1 p1)
           (list (setq e2 (entlast)) (getval 10 e2)) 0 ia)
(entdel e1) (entdel e2))

;----------------------------- DRAW PLINE ROUTINE ----------------------------|
;
;    SELECTION OF AXIS FOR REVSURF ON END (IE NORNAL) RESULTS IN INCONSISTANT
;    RESULTS.    ENTSEL LISTS ACT STRANGE.     SEE ALSO DRWARC.  NEED TO HANDLE
;    CLOSED PLINES.    AS WELL AS KEEP X-AXIS ALIGNED SUCH AS UCSXAXIS.

(defun drwpli (e / e1)
(while (/= (getval 0 (entnext e)) "SEQEND")
  (cond
     ((zerop (getval 42 e))  (drwlin e))          ;     LINE SEGMENT?
     (T (setq e1 e                                ;     ARC SEGMENT.
              e2 (entnext e1)
              p1 (getval 10 e1)                   ;     POINT 1
              p2 (getval 10 e2)                   ;     POINT 2
              b1 (getval 42 e1)                   ;     BULGE
              a1 (angle p1 p2)                    ;     ANGLE BETWEEN P1 & P2
              d1 (distance p1 p2)                 ;     DISTANCE BETWEEN P1 & P2
              ia (* 4 (atan b1))                  ;     INCLUDED ANGLE
              r1 (/ (/ d1 2) (sin (/ ia 2)))      ;     RADIUS
              ;  BUG POLAR 3D PT. SHOULD RETURN 3D POINT.
              c1 (append (polar p1 (+ a1 (- (/ pi 2) (/ ia 2))) r1)
                         (list (caddr p1)))       ;     center point
              c2 (mapcar '+ c1 '(0 0 -1)))
        (setvar "surftab1" (abs (fix (if (> (fix (/ (* 2 pi r1) arcdst)) arcmin)
                                       (/ (* ia r1) arcdst)
                                       (* arcmin (/ ia pi))))))
        (setq p1 (trans p1 path 0)
              p2 (trans p2 path 0)
              c1 (trans c1 path 0)
              c2 (trans c2 path 0))
        (command "ucs" "3p" p1 c1 c2
                 "insert" "*$$skin" '(0 0 0) 1 0
                 "ucs" "p")
        (setq e1 (entlast))
        (command "color" "1" "line" c1 c2 "" "color" "bylayer"
                 "revsurf" (list e1 p1)
                    (list (setq e2 (entlast)) c1) 0 (- ia))
        (entdel e1) (entdel e2)
     ))
; (while (/= (setq temp (getstring "\nVariable to test: ")) "")
;    (princ (strcat temp ": "))
;    (princ (eval (read temp)))
;    (command "line" (eval (read temp)) pause ""))
  (setq e (entnext e)))
;    SLOW SSGET FILTERS.  DEL SOME OTHER WAY.
(if (ssget "x" '((8 . "$$skin"))) (command "erase" "p" ""))
)

;--------------------------- DRAW SPLINE-FIT PLINE ---------------------------|
;
;    WORKS IN WCS POINTS.

(defun splseg ()
(command "ucs" "za" (midp p1 p2) p2
               "insert" "*$$skin" '(0 0 0) 1 0
               "ucs" "p"
               "rulesurf"
                  (list e2 '(0 0 0))
                  (list (setq e3 (entlast)) '(0 0 0)))
(entdel e2)
(setq p0 p1
      p1 p2
      e2 e3))

(defun drwspl (e / e1 e2 e3 p0 p1 p2)
(setvar "surftab1" (getvar "surftab2"))
(setq e1 (entnext e)  ;  FIRST POINT ALLWAYS CONTROL
      p1 (TRANS (getval 10 e1) PATH 0)
      p2 (TRANS (getval 10 (setq e1 (entnext e1))) PATH 0))
(command "ucs" "za" (midp p1 p2) p2
         "insert" "*$$skin" '(0 0 0) 1 0
         "ucs" "p")
(setq e2 (entlast))
(if closed
  (setq s1 (list p1 p2))
  (command "copy" e2 "" p2 (midp p1 p2)  ;  DO FIRST SEGMENT.
           "rulesurf"
              (list e2 '(0 0 0))
              (list (setq e3 (entlast)) '(0 0 0))))
(if e3 (entdel e3))
(setq p1 p2)
(while p2
  (setq p2 (TRANS (getval 10 (setq e1 (entnext e1))) PATH 0))
  (if (70bit? 16 e1) (setq p2 nil) (splseg))  ;  SPLINE FRAME CONTROL PT?
)
(if closed
  (progn
     (setq p2 (car s1))
     (splseg)
     (setq p2 (cadr s1))
     (entdel (splseg)))
  (command "copy" e3 "" (midp p0 p1) p1 ;  DO LAST SEGMENT.
           "rulesurf"
              (list e2 '(0 0 0))
              (list (setq e3 (entlast)) '(0 0 0))))
(entdel e3) (entdel e2))

;-------------------------------- MAIN PROGRAM -------------------------------|

(defun c:skin ();/ path patype notwcs outl closed)
(setq lisp-er *error*
      *error* skin-er)
(getset '(          ("ucsfollow"  0)  ("regenmode"  0)  ("expert"    2)
  ("cmdecho"    0)  ("snapmode"   0)  ("gridmode"   0)  ("blipmode"  1)
  ("highlight"  0)  ("aunits"     3)  ("ucsicon"    1)  ("flatland"  0)))
(cngset)
(cond
  ((setq path (getent '("POLYLINE" "LINE" "ARC") "path" nil))
     (setq patype   (getval 0 path)
           notwcs   (cond ((getval 210 path)) (T nil))
           outl     (outlin))
     (setvar "blipmode" 0) (command "ucs" "w")
     (cond
        ((eq patype      "ARC")    (drwarc path))
        ((eq patype     "LINE")    (drwlin path))
        ((eq patype "POLYLINE")                   ;  THEN SEND FIRST VERTEX.
           (setq closed (if (70bit? 1 path) T nil))
           (cond
              ((70bit? 16 path) (princ "\nCannot SKIN a 3DMESH."))
              ((70bit? 4  path) (drwspl (entnext path)))
              (T                (drwpli (entnext path)))))
        (T (princ "\nNot a defined POLYLINE TYPE."))))
  (T (princ "\nNo object found.")))
(command "undo" "end")
(moder)
(setq *error* lisp-er)
(if path (princ "\nProcess complete.") (princ "  Process terminated."))
(prin1))

;------------------------------------ END ------------------------------------|

(princ "Loaded. \n")
(prin1)
